import numpy as np
from itertools import combinations_with_replacement as cwr
from scipy.optimize import minimize as scimin
from scipy.stats import chi2 as chi2
from slicednormals import HyperEllipse as HE

class SNDist():
    "Ref: Improving the Uncertainty Quantifcation of Sliced Normal Distributions by Scaling the Covariance Matrix Colbert et al 2020"
    def __init__(self, Data, DoF):
        self.Data = Data
        self.DoF = DoF
        self.mu = None
        self.Sigma = None
    
    def SN_Phi(self, Data):
        # Unnormalised log-likelihood
        Z_Data = ZExpand(Data, self.DoF)
        return Phi(Z_Data, self.mu, self.Sigma)
    
class Basic_SN(SNDist):
    def __init__(self, Data, DoF):
        super().__init__(Data, DoF)
        self.Z_Data = ZExpand(self.Data, self.DoF)
        self.mu = np.mean(self.Z_Data,0)
        self.Sigma = Nearest_PSD(np.cov(self.Z_Data.T, ddof = 0))
    
class Scaled_SN(Basic_SN):
    def __init__(self, Data, DoF, No_Supp_Samples, Retain_Supp_Samples = False):
        super().__init__(Data, DoF)
        self.Hyper_Ellipse = HE.Hyper_Ellipse(Data = self.Data)
        self.Supp_Samples = self.Hyper_Ellipse.Sample(No_Supp_Samples)
        if Retain_Supp_Samples:
            self.Z_Supp_Samples = ZExpand(self.Supp_Samples, self.DoF)
            self.Z_Supp_Phi = -Phi(self.Z_Supp_Samples, self.mu, self.Sigma)
            self.Z_Supp_PhiSum = np.sum(np.exp(self.Z_Supp_Phi))
            self.Z_Data_PhiSum = np.sum(-Phi(self.Z_Data, self.mu, self.Sigma))
            objfun = lambda gamma: -((-np.shape(self.Data)[0] * np.log(np.sum(np.exp(self.Z_Supp_Phi * gamma)))) + gamma * self.Z_Data_PhiSum)
        else:
            Z_Supp_Samples = ZExpand(self.Supp_Samples, self.DoF)
            Z_Supp_Phi = -Phi(Z_Supp_Samples, self.mu, self.Sigma)
            Z_Data_PhiSum = np.sum(-Phi(self.Z_Data, self.mu, self.Sigma))
            objfun = lambda gamma: -((-np.shape(self.Data)[0] * np.log(np.sum(np.exp(Z_Supp_Phi * gamma)))) + gamma * Z_Data_PhiSum)
        self.gamma = scimin(objfun, 1e-9, bounds=((0, None),)).x
        self.Sigma = self.Sigma/self.gamma
        if Retain_Supp_Samples:
            self.Z_Supp_Phi = -Phi(self.Z_Supp_Samples, self.mu, self.Sigma)
            self.Z_Supp_PhiSum = np.sum(np.exp(self.Z_Supp_Phi))
            self.Z_Data_PhiSum = np.sum(-Phi(self.Z_Data, self.mu, self.Sigma))

def Phi(Data, mu, Sigma):
    # Unnormalised log-likelihood
    DMu = Data - mu
    return (np.sum(DMu.T*(np.linalg.solve(Sigma, DMu.T)),0)/2).T

def Nearest_PSD(Mat):
    try:
        np.linalg.cholesky(Mat)
        return Mat
    except np.linalg.LinAlgError:
        Symm = (Mat + Mat.T)/2
        [_, S, V] = np.linalg.svd(Symm)
        SymmPF = np.dot(V,np.dot(S,V.T))
        MatStar = (Mat + SymmPF)/2
        MatStar = (MatStar + MatStar.T)/2
        k = 0
        while True:
            try:
                np.linalg.cholesky(MatStar)
                break
            except np.linalg.LinAlgError:
                k += 1
                EMin = np.min(np.real(np.linalg.eig(MatStar)[0]))
                MatStar += (-EMin * k ** 2 + np.spacing(1)) * np.eye(np.shape(MatStar)[0])
        return MatStar

def ZExpand(P_Data, DoF = 1):
    Z_Data = P_Data
    Degree = 2
    while Degree <= DoF:
        if len(np.shape(P_Data)) > 1:
            Z_Data = np.hstack((Z_Data, np.vstack([[np.prod(a) for a in cwr(P_D, Degree)] for P_D in P_Data])))
        else:
            Z_Data = np.hstack((Z_Data, [np.prod(a) for a in cwr(P_Data, Degree)]))
        Degree += 1
    return Z_Data


import numpy as np
from scipy.optimize import minimize as scimin
from scipy.stats import chi2
from scipy.special import gamma
from numpy.random import default_rng


class Hyper_Ellipse():
    "Ref: Improving the Uncertainty Quantifcation of Sliced Normal Distributions by Scaling the Covariance Matrix Colbert et al 2020"
    def __init__(self, Data = None):
        self.Data = Data
        if not self.Data is None:
            self.Fit_Hyper_Ellipse()

    def Fit_Hyper_Ellipse(self, Data = None):
        """ Calculates hyper-ellipse parameters following following "Improving the Uncertainty Quantifcation of Sliced Normal Distributions by Scaling the Covariance Matrix" Colbert et al 2020 Appendix A. """
        if not Data is None:
            self.Data = Data
        mu = np.mean(self.Data, 0)
        Sigma = Nearest_PSD(np.cov(self.Data.T, ddof = 0))
        L = np.linalg.cholesky(Sigma)
        TempQ = np.linalg.solve(Sigma,mu.T)
        Q0 = Nearest_PSD(np.vstack((np.hstack((mu@TempQ, -(TempQ))),np.vstack((-TempQ, np.linalg.solve(L,np.linalg.solve(L.T,np.eye(np.shape(L)[0]))))).T)))
        Q0 = np.linalg.cholesky(Q0)
        HE_Q = scimin(self.HE_objfun, Q0).x
        l = int(len(HE_Q)**0.5)
        HE_Q = np.reshape(HE_Q,(l,l))
        self.Q = HE_Q@HE_Q.T

        self.P_Matrix = self.Q[1:,1:]
        self.Chol_P_Matrix = np.linalg.cholesky(Nearest_PSD(self.P_Matrix))
        self.Sigma = np.linalg.solve(self.Chol_P_Matrix, np.linalg.solve(self.Chol_P_Matrix.T, np.eye(np.shape(self.Chol_P_Matrix)[0])))
        self.mu = np.linalg.solve(-self.P_Matrix, self.Q[1:, 0])
        self.Beta = self.FBeta(self.Q)
        self.Dims = np.shape(self.Data)[1]
        DMu = self.Data[np.argmax(np.linalg.norm(self.Data-self.mu, axis=1)),:] - self.mu
        self.Threshold = np.min(np.sum(DMu.T*(np.linalg.solve(self.Sigma, DMu.T)),0))*3
        self.Volume = (np.pi**(self.Dims/2)/gamma(1+self.Dims/2))*(np.linalg.det(self.Sigma)**0.5)*(self.Threshold**(self.Dims/2))

    def HE_objfun(self, Q):
        l = int(len(Q)**0.5)
        Q = np.reshape(Q,(l,l))
        Q = Q@Q.T
        return -((2*np.sum(np.log(np.diag(np.linalg.cholesky(Nearest_PSD(Q[1:,1:]))))))+self.FBeta(Q))

    def FBeta(self, Q):
        Beta = 1
        Beta = np.min(np.sum(-np.column_stack([np.ones(np.shape(self.Data)[0]),self.Data]).T*(Q@np.column_stack([np.ones(np.shape(self.Data)[0]),self.Data]).T)/2, 0))
        return Beta

    def Sample(self, No_Samples = None):
        return Dezert_Musso(self.Threshold, self.Sigma, self.mu, No_Samples)

def Dezert_Musso(Threshold,Sigma,mu,No_Samples):
    rng = default_rng()
    Dims = np.shape(Sigma)[0]

    HSphere = rng.standard_normal(size=(No_Samples, Dims))
    HSphere = HSphere.T/np.sqrt(np.sum(HSphere**2,1))

    Counter = rng.random(size=No_Samples)**(1/Dims)
    unif_sph = Counter*HSphere
    Chol_S = np.linalg.cholesky(Sigma)
    unif_ell = Chol_S.T @ unif_sph
    return (unif_ell*np.sqrt(Threshold)+np.array([mu]).T).T

def Nearest_PSD(Mat):
    try:
        np.linalg.cholesky(Mat)
        return Mat
    except np.linalg.LinAlgError:
        Symm = (Mat + Mat.T)/2
        [_, S, V] = np.linalg.svd(Symm)
        SymmPF = np.dot(V,np.dot(S,V.T))
        MatStar = (Mat + SymmPF)/2
        MatStar = (MatStar + MatStar.T)/2
        k = 0
        while True:
            try:
                np.linalg.cholesky(MatStar)
                break
            except np.linalg.LinAlgError:
                k += 1
                EMin = np.min(np.real(np.linalg.eig(MatStar)[0]))
                MatStar += (-EMin * k ** 2 + np.spacing(1)) * np.eye(np.shape(MatStar)[0])
        return MatStar


from setuptools import setup
setup(
  name = 'slicednormals',
  packages = ['slicednormals'],
  version = '0.1.6-alpha',
  license='gpl-3.0',
  description = 'Library for fitting basic and scaled sliced normal distributions',
  author = 'Alexander Wimbush',
  author_email = 'alexanderpwimbush@gmail.com', 
  url = 'https://github.com/Institute-for-Risk-and-Uncertainty/Sliced-Normals-Python',
  download_url = 'https://github.com/Institute-for-Risk-and-Uncertainty/Sliced-Normals-Python/archive/0.1.6-alpha.tar.gz', 
  keywords = ['dependence', 'expansion', 'density'], 
  install_requires=[
          'numpy',
		  'scipy',
      ],
)